/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.comparators.SparseLocalDerivative.SparseDeriativeEntry;
import de.unibi.citec.tcs.alignment.comparators.SparseLocalDerivative.TrivialSparseLocalDerivative;
import de.unibi.citec.tcs.alignment.sequence.ValueType;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;
import java.util.Iterator;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class ComparisonBasedSkipExtendedComparatorTest {

	public ComparisonBasedSkipExtendedComparatorTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of the static version, which is basically a SkipExtendedComparator.
	 */
	@Test
	public void testStatic() {
		final TestImpl instance = new TestImpl();
		//first do the trivial check for compare (which is manually implemented below).
		assertEquals(0, instance.compare(new VectorialValue(1), new VectorialValue(1)), 0);
		assertEquals(instance.alpha, instance.compare(new VectorialValue(0), new VectorialValue(1)), 0);
		//now check the other modes.
		final double expectedDeletion = 0.1;
		instance.setDeletionCost(expectedDeletion);
		assertEquals(expectedDeletion, instance.delete(new VectorialValue(1)), 0);
		assertEquals(expectedDeletion, instance.delete(new VectorialValue(7)), 0);

		final double expectedInsertion = 0.2;
		instance.setInsertionCost(expectedInsertion);
		assertEquals(expectedInsertion, instance.insert(new VectorialValue(1)), 0);
		assertEquals(expectedInsertion, instance.insert(new VectorialValue(7)), 0);

		final double expectedSkipDeletion = 0.3;
		instance.setSkipDeletionCost(expectedSkipDeletion);
		assertEquals(expectedSkipDeletion, instance.skipDelete(new VectorialValue(1)), 0);
		assertEquals(expectedSkipDeletion, instance.skipDelete(new VectorialValue(7)), 0);

		final double expectedSkipInsertion = 0.4;
		instance.setSkipInsertionCost(expectedSkipInsertion);
		assertEquals(expectedSkipInsertion, instance.skipInsert(new VectorialValue(1)), 0);
		assertEquals(expectedSkipInsertion, instance.skipInsert(new VectorialValue(7)), 0);

		//check derivatives.
		assertEquals(5, instance.getNumberOfParameters());
		assertEquals(1, instance.calculateLocalDerivative(0, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(4, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		Iterator<SparseDeriativeEntry> sparseDerivative
				= instance.calculateSparseLocalDerivative(
						new VectorialValue(1), null, OperationType.DELETION).iterator();
		assertTrue(sparseDerivative.hasNext());
		SparseDeriativeEntry next = sparseDerivative.next();
		assertEquals(0, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(1, instance.calculateLocalDerivative(1, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(4, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(null,
				new VectorialValue(1), OperationType.INSERTION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(1, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(1, instance.calculateLocalDerivative(2, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(4, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(
				new VectorialValue(1), null, OperationType.SKIPDELETION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(2, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(1, instance.calculateLocalDerivative(3, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(4, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(null,
				new VectorialValue(1), OperationType.SKIPINSERTION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(3, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(1,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(2,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(3,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(1, instance.calculateLocalDerivative(4,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0.5, instance.calculateLocalDerivative(4,
				new VectorialValue(0), new VectorialValue(0.5),
				OperationType.REPLACEMENT), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

	}

	/**
	 * Test of the dynamic version.
	 */
	@Test
	public void testDynamic() {
		final TestImpl instance = new TestImpl();
		instance.setDynamic(true);
		instance.alpha = 1;
		//first do the trivial check for compare (which is manually implemented below).
		assertEquals(0, instance.compare(new VectorialValue(1), new VectorialValue(1)), 0);
		assertEquals(instance.alpha, instance.compare(new VectorialValue(0), new VectorialValue(1)), 0);
		//now check the other modes.
		assertEquals(1, instance.delete(new VectorialValue(1)), 0);
		assertEquals(7, instance.delete(new VectorialValue(7)), 0);

		assertEquals(1, instance.insert(new VectorialValue(1)), 0);
		assertEquals(5, instance.insert(new VectorialValue(5)), 0);

		assertEquals(instance.getSkipFactor(), instance.skipDelete(new VectorialValue(1)), 0);
		assertEquals(7 * instance.getSkipFactor(), instance.skipDelete(new VectorialValue(7)), 0);

		assertEquals(instance.getSkipFactor(), instance.skipInsert(new VectorialValue(1)), 0);
		assertEquals(9 * instance.getSkipFactor(), instance.skipInsert(new VectorialValue(9)), 0);

		//check derivatives.
		assertEquals(5, instance.getNumberOfParameters());
		assertEquals(0, instance.calculateLocalDerivative(0, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		assertEquals(1, instance.calculateLocalDerivative(4, new VectorialValue(1),
				null, OperationType.DELETION), 0);
		Iterator<SparseDeriativeEntry> sparseDerivative
				= instance.calculateSparseLocalDerivative(
						new VectorialValue(1), null, OperationType.DELETION).iterator();
		assertTrue(sparseDerivative.hasNext());
		SparseDeriativeEntry next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		assertEquals(1, instance.calculateLocalDerivative(4, null,
				new VectorialValue(1), OperationType.INSERTION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(null,
				new VectorialValue(1), OperationType.INSERTION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		assertEquals(instance.getSkipFactor(), instance.calculateLocalDerivative(4, new VectorialValue(1),
				null, OperationType.SKIPDELETION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(
				new VectorialValue(1), null, OperationType.SKIPDELETION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(instance.getSkipFactor(), next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(1, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(2, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(0, instance.calculateLocalDerivative(3, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		assertEquals(instance.getSkipFactor(), instance.calculateLocalDerivative(4, null,
				new VectorialValue(1), OperationType.SKIPINSERTION), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(null,
				new VectorialValue(1), OperationType.SKIPINSERTION).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(instance.getSkipFactor(), next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());

		assertEquals(0, instance.calculateLocalDerivative(0,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(1,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(2,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0, instance.calculateLocalDerivative(3,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(1, instance.calculateLocalDerivative(4,
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT), 0);
		assertEquals(0.5, instance.calculateLocalDerivative(4,
				new VectorialValue(0), new VectorialValue(0.5),
				OperationType.REPLACEMENT), 0);
		sparseDerivative = instance.calculateSparseLocalDerivative(
				new VectorialValue(0), new VectorialValue(1),
				OperationType.REPLACEMENT).iterator();
		assertTrue(sparseDerivative.hasNext());
		next = sparseDerivative.next();
		assertEquals(4, next.getParameterIndex());
		assertEquals(1, next.getDerivative(), 0);
		assertFalse(sparseDerivative.hasNext());
	}

	/*
	 * This is basically an easy L1-norm comparator that just works for
	 * one-dimensional input, meaning: it computes the absolute value of the
	 * difference of two scalars: |a - b|
	 *
	 * To make it somewhat more interesting this value is then multiplied by
	 * a parameter alpha, which can be learned.
	 */
	public static class TestImpl extends ComparisonBasedSkipExtendedComparator<VectorialValue> {

		double alpha = 0.1;

		public TestImpl() {
			super(new VectorialValue(0), new VectorialValue(0),
					new VectorialValue(0), new VectorialValue(0));
		}

		@Override
		public double comparisonDerivative(int paramIdx, VectorialValue a, VectorialValue b) {
			if (paramIdx != 4) {
				return 0;
			}
			double val = Math.abs(a.getVector()[0] - b.getVector()[0]);
			return val;
		}

		@Override
		public SparseLocalDerivative sparseComparisonDerivative(VectorialValue a, VectorialValue b) {
			final TrivialSparseLocalDerivative out = new TrivialSparseLocalDerivative();
			double val = Math.abs(a.getVector()[0] - b.getVector()[0]);
			out.add(4, val);
			return out;
		}

		@Override
		public int getNumberOfComparisonParameters() {
			return 1;
		}

		@Override
		public ValueType getType() {
			return ValueType.VECTOR;
		}

		@Override
		public double compare(VectorialValue a, VectorialValue b) {
			return alpha * Math.abs(a.getVector()[0] - b.getVector()[0]);
		}
	}

}
